;; Vraca element "elem" ako je trazene vrste "vrsta", inace nil
(defun vrsta-elementa (elem vrsta / entd)
  (if elem
    (progn
      (setq entd (entget (car elem)))
      (if (member (cdr (assoc 0 entd)) vrsta)
	elem
	nil))
    nil))

;; Funkcije za jednostavne vektorske operacije u 3D, ovo sigurno vec postoji
;; valjalo bi da rade i sa vise od dva clana, kao sto rade i + i -, ali nije preterano bitno
(defun vec+ (p1 p2)
  (list (+ (car p1) (car p2))
	(+ (cadr p1) (cadr p2))
	(+ (caddr p1) (caddr p2))))

(defun vec- (p1 p2)
  (list (- (car p1) (car p2))
	(- (cadr p1) (cadr p2))
	(- (caddr p1) (caddr p2))))

(defun vec* (k p1)
  (list (* k (car p1))
	(* k (cadr p1))
	(* k (caddr p1))))


;; Izbor tacke
(defun dajtacku (text)
  (vrsta-elementa (entsel text) '("POINT")))

(defun postavi-ime-sloja ()
  (setq ime-sloja (getstring "\nUnesite ime sloja: ")))

;; Postavljanje novih koordinata za tacku
(defun nove-koordinate (ent koord / entd)
  ;; dobija "entitet" i nove koordinate
  (if (not ime-sloja) (postavi-ime-sloja))
  (setq entd (entget ent))
  (setq entd (subst (cons 8 ime-sloja) (assoc 8 entd) (subst (cons 10 koord) (assoc 10 entd) entd)))
  (entmake entd)
  ;(entupd ent)
  entd)

(defun c:digni-tacku (/ pt entd koord rast)
  ; Trenutno podrzavamo samo tacke ("POINT"), sa "LINE" i slicno mozda naknadno?
  ; (ako umesto dajtacku koristimo (vrsta-elementa ... '("POINT" "LINE"))
  ; onda ce se "dignuti" prva tacka linije, a NE izabrana [sto bi bilo ispravno] --
  ; dok ne vidim kako se moze saznati koja je tacka izabrana, ostaje ovako)
  (if (setq pt (dajtacku "\nIzaberi tacku:"))
    (progn
  	(setq entd (entget (car pt)))
        (setq koord (cdr (assoc 10 entd)))
        (initget 1 "Neg")
        (setq rast (getdist "\nVisina (Neg): "))
        (if (eq rast "Neg")
	  (progn
	    (setq rast (- 0 (getdist "\nVisina: ")))))
  	(setq koord (list (car koord) (cadr koord) (+ rast (caddr koord))))
	(nove-koordinate (car pt) koord)) ;; vratimo nove koordinate
    (alert "Morate izabrati tacku.")))

;; Pokrenuti po ucitavanju centralne projekcije
;; centar je globalna promenljiva
(defun izaberi-centar (/ cent)
  (if (setq cent (vrsta-elementa (entsel "\nIzaberite centar projekcije: ") '("POINT")))
    (setq centar (cdr (assoc 10 (entget (car cent)))))
    (setq centar nil)))

;; Pozvati kada zelimo da promenimo centar projektovanja
(defun c:promeni-centar ()
  (izaberi-centar))

;; Prava-nosac je odredjena sa dve tacke: tragom i nedogledom
;; Ravan-nosac je odredjena sa dve prave: tragom i nedoglednicom
;; Ukoliko nije sacinjen potpun izbor, vracamo "nil" (da bismo koristili u ciklusu)
(defun izaberi-nosac (/ prvo drugo a b c d)
    (if (setq prvo (vrsta-elementa (entsel "\nIzaberite trag (tacka/prava): ") '("POINT" "LINE" "XLINE")))
      (if (vrsta-elementa prvo '("POINT"))
	;; radi se o pravoj-nosacu, treba nam jos jedna tacka
       (if (setq drugo (vrsta-elementa (entsel "\nIzaberite nedogled: ") '("POINT")))
	 ;; imamo dve tacke, vratimo ih
	 (list 'prava (cdr (assoc 10 (entget (car prvo)))) (cdr (assoc 10 (entget (car drugo)))))
	 nil)
	(progn ;; inace imamo pravu, daj jos jednu pravu
          (setq drugo (vrsta-elementa (entsel "\nIzaberite nedoglednicu: ") '("LINE", "XLINE")))
	  (setq a (cdr (assoc 10 (entget (car prvo))))
		b (cdr (assoc 11 (entget (car prvo))))
		c (cdr (assoc 10 (entget (car drugo))))
		d (cdr (assoc 11 (entget (car drugo)))))
	  (princ (list a b c d))
	  (if (vrsta-elementa prvo '("XLINE"))
	      (setq prva (list a (vec+ a b))
		    druga (list c (vec+ c d)))
	      (setq prva (list a b)
		    druga (list c d)))
	  (list 'ravan prva druga)))
      nil))

(defun nadji-najblizu-tacku (koord / rast tmprast el rez)
  (setq el (entnext))
  (setq rez nil)
  (setq rast 20.0) ; necemo veca rastojanja
  (while el
    (setq ent (entget el))
    (if (equal (cdr (assoc 0 ent)) "POINT")
	(if (< (setq tmprast (distance koord (cdr (assoc 10 ent)))) rast)
	    (setq rast tmprast
		  rez el)))
    (setq el (entnext el)))
  rez)

(defun c:proba-izbora-tacke (/)
  (setq ss (ssadd)) ; prazan izbor
  (if (setq el (nadji-najblizu-tacku (cadr (entsel "\nIzaberite tacku: "))))
      (progn (ssadd el ss)
	     (sssetfirst nil ss)))
  (setq ss nil))
  

;; Da li su tacke a i b sa iste strane prave t0--t1 (zapravo, da li se
;; duzi a--b i t0+(t0-t1)--t1+(t1-t0) *ne* seku)
;; (radi samo za tacke u istoj ravni)
(defun sa-iste-strane (a b t0 t1)
  (if (inters a b (vec+ t0 (vec- t0 t1)) (vec+ t1 (vec- t1 t0)))
    nil t))

(defun c:odprojektuj-centralno (/ nosac tacka enttacka trag nedg q tmp ptrag pnedg koord)
  (if (not centar) (setq centar (izaberi-centar)))
  (if centar
    (progn
      (while (setq nosac (izaberi-nosac))
        (progn
          (princ "Nosac: ") (princ nosac)
          (while (setq tacka (vrsta-elementa (entsel "\nIzaberite tacku: ") '("POINT")))
            (progn
	      (setq enttacka (entget (car tacka)))
	      (setq vrsta (car nosac))
	      (if (eq vrsta 'prava)
		(progn
		  (setq trag (cadr nosac)
			nedg (caddr nosac)))
		(progn ; inace ravan
		  (setq ptrag (cadr nosac)
			pnedg (caddr nosac)
		        ;; neka tacka sa nedoglednice, npr. prva
	                nedg (car pnedg))
		  (princ "ptrag: ")(princ ptrag)
		  ;; presek traga i prave koja sadrzi projekciju tacke i neku tacku sa nedoglednice
		  (setq trag (inters nedg (cdr (assoc 10 enttacka)) (car ptrag) (cadr ptrag) nil))))
	      (setq tmp (vec+ centar (vec- trag nedg)))
	      (princ "\ntrag: ")(princ trag)
	      (princ "\nnedg: ")(princ nedg)
	      (princ "\ntmp: ")(princ tmp)
	      (princ "\ncentar: ")(princ centar)
              ;; zbog loseg zaokrugljivanja u AutoCAD-u, ovako ne moze:
	      (princ (setq koord (inters trag tmp centar (cdr (assoc 10 enttacka)) nil))))
	      (nove-koordinate (car tacka) koord))))
      'gotovo) ;; da li mozda vratiti broj izmenjenih tacaka, ili ne vratiti nista?
    (alert "Nepoznat centar projekcije!")))

(defun jv-dodajtacku (tacka / poz)
  (if (not (setq poz (vl-position tacka jv-tacke)))
    (progn
      (setq jv-tacke (append jv-tacke (list tacka)))
      (setq poz (vl-position tacka jv-tacke))))
  (princ poz))

(defun c:jv-ocistisve ()
  (setq jv-tacke nil
	jv-povrsi nil))

(defun c:jv-povrs (/ tacka enttacka povrs tacke)
  (setq povrs nil) ; zapocni novu povrs
  (setq tacke nil) ; lista tacaka za 3dface komandu
  (while (setq tacka (vrsta-elementa (entsel "\nIzaberite tacku: ") '("POINT")))
    (progn
      (setq enttacka (entget (car tacka)))
      (setq brojtacke (jv-dodajtacku (cdr (assoc 10 enttacka))))
      (setq tacke (append tacke (list (cdr (assoc 10 enttacka)))))
      ;(princ "\nbroj tacke:")(princ brojtacke)
      (setq povrs (append povrs (list brojtacke)))))
  (if (and povrs (> (length povrs) 1)) 
      (progn
	(setq jv-povrsi (append jv-povrsi (list povrs)))
	(if (> (length povrs) 4)
	; vise od 4 tacke
	    (alert "Naredba 3dface ne podrzava vise od 4 tacke za povrsi!\nPovrs nece biti iscrtana u AutoCAD-u.")
	(if (> (length povrs) 3) 
	; 4 razlicite tacke
	    (command "3dface" (car tacke) (cadr tacke) (caddr tacke) (cadddr tacke) nil)
	(if (> (length povrs) 2)
	; 3 razne tacke 
	    (command "3dface" (car tacke) (cadr tacke) (caddr tacke) (caddr tacke) nil))))))
  tacke)

(defun jv-dajravan (/ svetacke tacka maxx maxy minx miny prosirenje)
  (setq prosirenje 0.15) ; koliko "prosirujemo" dobijenu ravan
  (setq svetacke jv-tacke
        tacka (car jv-tacke))
  (setq minx (car tacka)
	maxx (car tacka)
	miny (cadr tacka)
	maxy (cadr tacka))
  (while (setq tacka (car svetacke))
    ;(princ tacka)(princ "\n")
    (if (< (car tacka) minx) 
	(setq minx (car tacka)))
    (if (> (car tacka) maxx) 
	(setq maxx (car tacka)))
    (if (< (cadr tacka) miny) 
	(setq miny (cadr tacka)))
    (if (> (cadr tacka) maxy) 
	(setq maxy (cadr tacka)))
    (setq svetacke (cdr svetacke)))
  (list (- minx (* (- maxx minx) prosirenje)) 
	(- miny (* (- maxy miny) prosirenje)) 
	(+ maxx (* (- maxx minx) prosirenje)) 
	(+ maxy (* (- maxy miny) prosirenje))))

(defun jv-saraznihstrana (tacka1 tacka2)
  (< (* (caddr (nth tacka1 jv-tacke)) (caddr (nth tacka2 jv-tacke))) 0))

(defun jv-iznadravni (tacka1)
  (> (caddr (nth tacka1 jv-tacke)) 0))

(defun jv-medjutacka (tacka1 tacka2 / koord1 koord2 x y z)
  (setq koord1 (nth tacka1 jv-tacke)
	koord2 (nth tacka2 jv-tacke))
  (if (caddr koord2)
      (setq odnos (- (/ (caddr koord1) (caddr koord2)))
	    x     (/ (+ (car koord1) (* odnos (car koord2))) (+ 1 odnos))
	    y     (/ (+ (cadr koord1) (* odnos (cadr koord2))) (+ 1 odnos))
	    z     0)
      (setq x (car koord2)
	    y (cadr koord2)
	    z (caddr koord2)))
  (jv-dodajtacku (list x y z))) ;; vraca broj medju-tacke

(defun jv-podelipovrsi (in-povrsi / povrsi iznad ispod prethodna tacka)
  (setq povrs (car in-povrsi))
  (if povrs
      (progn
	(setq	prethodna (car povrs)
		preseka 0)
	(if (jv-iznadravni prethodna)
	    (setq iznad (list prethodna)
		  ispod nil)
	  (setq iznad nil
		ispod (list prethodna)))
	(foreach tacka (cdr povrs) (progn
	  (if (jv-saraznihstrana prethodna tacka)
	      (setq iznad (append iznad (list (jv-medjutacka prethodna tacka)))
		    ispod (append ispod (list (jv-medjutacka prethodna tacka)))
		    preseka (+ 1 preseka)))
	  (if (jv-iznadravni tacka)
	      (setq iznad (append iznad (list tacka)))
	    (setq ispod (append ispod (list tacka))))))
	(if (and iznad ispod (= preseka 2)) ; delili smo povrs
	    (append (list iznad ispod) (jv-podelipovrsi (cdr in-povrsi)))
	  (append (list povrs) (jv-podelipovrsi (cdr in-povrsi)))))
    ; ako nema povrsi, vrati nil
    nil))
  
	   

(defun jv-izvoz-projektivne-ravni (/ projektivna-ravan povrs ravan tacka1 tacka2 tacka3 tacka4)
  (initget "Da Ne")
  (setq projektivna-ravan (getkword "Zelite li da izvezete i projektivnu ravan (Da/Ne)? "))
  (if (and (not jv-ravan-izvezena) (eq projektivna-ravan "Da"))
      (progn
	(setq ravan (jv-dajravan))
 	(setq tacka1 (jv-dodajtacku (list (car ravan) (cadr ravan) 0.0))
	      tacka2 (jv-dodajtacku (list (caddr ravan) (cadr ravan) 0.0))
	      tacka3 (jv-dodajtacku (list (caddr ravan) (nth 3 ravan) 0.0))
	      tacka4 (jv-dodajtacku (list (car ravan) (nth 3 ravan) 0.0))
	      povrs (list tacka1 tacka2 tacka3 tacka4))
	(setq jv-povrsi (append (jv-podelipovrsi jv-povrsi) (list povrs)))
	(setq jv-ravan-izvezena t))
    nil))
    

(defun jv-izborfajla ()
  (getstring "Unesite ime fajla: "))

(defun c:jv-pravifajl (/ f tacka strana broj)
  (jv-izvoz-projektivne-ravni)
  (setq f (open (jv-izborfajla) "w"))
  (if f
      (progn
	; dodaj centar projekcije ako je definisan
	(if centar (jv-dodajtacku centar))
	; prvo ispis tacka
	(while (setq tacka (car jv-tacke))
	  (princ "v " f) (princ (car tacka) f)
	  (princ " " f) (princ (cadr tacka) f)
	  (princ " " f) (princ (caddr tacka) f)
	  (princ "\n" f)
	  (setq jv-tacke (cdr jv-tacke)))
	; pa ispis strana
	(while (setq strana (car jv-povrsi))
	  ;(princ "\nstrana: ")(princ strana)
	  (princ "f" f)
	  (while (setq broj (car strana))
	    (princ " " f) (princ (+ broj 1) f)
	    (setq strana (cdr strana)))
	  (princ "\n" f)
	  (setq jv-povrsi (cdr jv-povrsi)))
	; gotovo, zatvori fajl
	(close f)
	nil)))
